import os
import shutil
import time
from pathlib import Path

import usernetes.utils as utils
from usernetes.config import ComposeConfig
from usernetes.logger import logger


class UsernetesRunner:
    """
    A Usernetes Runner will run usernetes via docker compose.

    We do this easily by using the docker python SDK. Note that
    we don't have an equivalent to "docker compose render"
    """

    def __init__(
        self,
        container_engine="docker",
        workdir=None,
        wait_seconds=5,
        compose_file="docker-compose.yaml",
    ):
        """
        Create a new transformer backend, accepting any options type.

        Validation of transformers is done by the registry
        """
        # Set and validate the working directory
        self.compose_file = compose_file
        self.set_workdir(workdir)
        self.compose = ComposeConfig(container_engine)
        # Break buffer time between running commands
        self.sleep = wait_seconds
        # Prepare a filesystem cache for workers to indicate readiness
        self.prepare_worker_cache()

    @property
    def usernetes_uid(self):
        """
        The uid for different assets (network, node) can be assembled
        from the basename of the working directory. E.g.,

           node: usernetes-7fnptyat-node
        network: usernetes-7fnptyat_default
        """
        return os.path.basename(self.workdir)

    def set_workdir(self, workdir):
        """
        Set and validate that the working directory exists.

        It must be populated with a usernetes clone or release, and have
        docker-compose.yaml.
        """
        self.workdir = workdir or os.getcwd()
        if not os.path.exists(self.workdir):
            raise ValueError(f"Working directory with usernetes {self.workdir} does not exist.")

        # This assumes usernetes has a docker-compose.yaml
        compose_file = os.path.join(self.workdir, self.compose_file)
        if not os.path.exists(compose_file):
            raise ValueError(f"{self.compose_file} does not exist in {self.workdir}.")

    def prepare_worker_cache(self):
        """
        This can be improved upon, but we need to sync external ips on the control
        plane after workers are ready. To do that, we prepare a directory where
        they will write their hostnames. The lead thus needs to know how many workers
        to expect. This should be created by the control plane, brought up first.
        """
        self.worker_cache = os.path.join(self.workdir, "worker-ready-cache")
        if not os.path.exists(self.worker_cache):
            logger.debug(f"Creating worker cache {self.worker_cache}")
            os.makedirs(self.worker_cache)

    def start_control_plane(self, worker_count, serial=False):
        """
        Start the usernetes control plane.

        This currently starts with the working directory as specified by
        the user. In practice, it makes sense for an admin to clone usernetes
        and stage this in a temporary location. The worker count is needed to
        determine when all workers are ready for a final sync.
        """
        self.up()
        # Note this was originally 10
        time.sleep(self.sleep)
        self.kubeadm_init()
        time.sleep(self.sleep)
        self.install_flannel()

        # Generate <self.workdir>/kubeconfig and join-command
        # As soon as this is generated, workers will start preparing
        self.ensure_kubeconfig()
        self.join_command()

        # We don't print anything because it's printed in the interface for the user
        # Serial mode usually means we need to manually move the join-command before
        # we issue the final sync to the external ips.
        if serial:
            logger.debug("Running in serial mode, returning early")
            return

        # Next, we need the workers to ready, and we sync the external IP once more
        self.wait_for_workers(worker_count)
        self.sync_external_ip()

    def clean(self, cache_only=True, keep_nodes=False):
        """
        Clean removes the usernetes root and all assets, unless cache only is
        set to true, in which case we just remove the worker cache. It also
        stops and removes the nodes, assuming we want to re-create them later.
        """
        # You can't remove everything but ask to keep nodes
        if not cache_only and keep_nodes:
            raise ValueError("To keep nodes, you must not delete the working directory.")
        if os.path.exists(self.worker_cache):
            logger.debug(f"Cleaning up worker cache {self.worker_cache}")
            shutil.rmtree(self.worker_cache)
        if cache_only:
            return
        if os.path.exists(self.workdir):
            logger.debug(f"Cleaning up usernetes root {self.workdir}")
            shutil.rmtree(self.workdir)
        if keep_nodes:
            return

        # Final cleanup of node (stop and remove) and network.
        self.cleanup_node()

    def cleanup_node(self):
        """
        Cleanup the node, including stop/remove of the image and network.
        """
        self.stop()
        # This is rm, not rmi
        self.remove()
        self.remove_network()

    def wait_for_workers(self, worker_count):
        """
        Wait for workers to indicate ready by writing their hostname
        """
        while True:
            print(f"Waiting for {worker_count} workers to be ready...")
            time.sleep(self.sleep)
            count = len(os.listdir(self.worker_cache))

            # The workers are ready, break from waiting
            if count == worker_count:
                print(f"⭐ Workers (N={worker_count}) are ready!")
                time.sleep(self.sleep)
                break

    def wait_for_control_plane(self):
        """
        Wait for the control plane to be ready.

        This is indicated by the presence of the join-command in the usernetes
        root. It is the responsibility of the executor / control plane to get it
        there. For shared filesystems, this should not be an issue.
        """
        join_command = os.path.join(self.workdir, "join-command")
        while True:
            print(f"Waiting for join-command in {self.workdir}...")
            time.sleep(self.sleep)
            if os.path.exists(join_command):
                print(f"⭐ Found join-command in {self.workdir}!")
                return

    def start_worker(self):
        """
        Start a usernetes worker (kubelet)
        """
        self.up()
        self.wait_for_control_plane()
        self.kubeadm_join()

        # Indicate we are ready!
        ready_path = os.path.join(
            self.worker_cache, f"{self.compose.usernetes_node_name}.ready.txt"
        )
        Path(ready_path).touch()

    @property
    def kubeconfig(self):
        """
        Path to the kubeconfig in the working directory
        """
        kubeconfig = os.path.join(self.workdir, "kubeconfig")
        os.environ["KUBECONFIG"] = kubeconfig
        os.putenv("KUBECONFIG", kubeconfig)
        return kubeconfig

    def ensure_kubeconfig(self):
        """
        Generate kubeconfig locally
        """
        if os.path.exists(self.kubeconfig):
            return
        # This will generate "kubeconfig" in self.workdir
        self.run_command(["make", "kubeconfig"])

    def join_command(self):
        """
        Generate the join-command (should be run by control plane)

        @echo "# Copy the 'join-command' file to another host, and run the following commands:"
            @echo "# On the other host (the new worker):"
            @echo "#   make kubeadm-join"
            @echo "# On this host (the control plane):"
            @echo "#   make sync-external-ip"
        """
        # kubeadm token create --print-join-command | tr -d '\r'
        self.run_command(["make", "join-command"])

    def run_command(self, command, do_check=True, success_code=0, quiet=False, allow_fail=False):
        """
        Wrapper to utils.run_command, and assumed in the working directory.

        I was originally re-creating the usernetes logic, but this is easier
        to maintain (less likely to break when the logic changes).
        """
        with utils.workdir(self.workdir):
            logger.debug(" ".join(command))
            result = utils.run_command(command, stream=True)

        # Assume we don't need to return the return code
        # can change if needed
        if do_check and result["return_code"] != success_code:
            # Allow to fail
            if not allow_fail:
                command = " ".join(command)
                raise ValueError(f"Issue running {command}: {result['return_code']}")
            msg = f"{command} ({result['return_code']})"
            print("Warning: issue running {msg} but allow fail is true.")
        response = result["message"]
        if response is not None and not quiet:
            print(response)
        return response

    def kubeadm_init(self):
        """
        kubeadm init
        """
        # $(NODE_SHELL) sh -euc "envsubst </usernetes/kubeadm-config.yaml >/tmp/kubeadm-config.yaml"
        # $(NODE_SHELL) kubeadm init --config /tmp/kubeadm-config.yaml --skip-token-print
        self.run_command(["make", "kubeadm-init"])

    def sync_external_ip(self):
        self.run_command(["make", "sync-external-ip"])

    def kubeadm_join(self):
        """
        kubeadm join
        """
        self.run_command(["make", "kubeadm-join"])

    def kubeadm_reset(self):
        """
        kubeadm reset
        """
        self.run_command(["make", "kubeadm-reset"])

    def install_flannel(self):
        """
        Install flannel networking fabric
        """
        self.run_command(["make", "install-flannel"])

    def get_pods(self):
        """
        Get pods with kubectl
        """
        utils.run_command(["kubectl", "get", "pods", "-A"], envars={"KUBECONFIG": self.kubeconfig})

    def debug(self):
        """
            @echo '# Debug'
            @echo 'make logs'
        @echo 'make shell'
            @echo 'make kubeadm-reset'
            @echo 'make down-v'
        @echo 'kubectl taint nodes --all node-role.kubernetes.io/control-plane-'
        """
        self.run_command(["make", "debug"])

    def up(self):
        """
        Run docker-compose up, always with detached.
        """
        with utils.workdir(self.workdir):
            self.compose.check()

            # Set needed environment variables
            self.compose.set_build_environment()

            # $(COMPOSE) up --build -d
            self.run_command(["make", "up"])

    def down(self, verbose=True):
        """
        Run docker-compose down.
        """
        if verbose:
            return self.run_command(["make", "down"])
        return self.run_command(["make", "down-v"])

    def stop(self, allow_fail=True):
        """
        Run docker stop - not supported in the Makefile
        """
        node = f"{self.usernetes_uid}-node-1"
        self.run_command(["docker", "stop", node], allow_fail=allow_fail)

    def remove_image(self, allow_fail=True):
        """
        Remove the node image.
        """
        node = f"{self.usernetes_uid}-node-1"
        self.run_command(["docker", "rmi", node], allow_fail=allow_fail)

    def remove_network(self, allow_fail=False):
        """
        Remove the node image.
        """
        network = f"{self.usernetes_uid}_default"
        self.run_command(["docker", "network", "rm", network], allow_fail=allow_fail)

    def remove(self):
        """
        Remove the node, and the network.
        """
        node = f"{self.usernetes_uid}-node"
        self.run_command(["docker", "rm", node])

    def logs(self):
        """
        Get logs from journalctl
        """
        self.run_command(["make", "logs"])

    def shell(self):
        """
        Execute a shell to the container.
        """
        with utils.workdir(self.workdir):
            os.system("make shell")



from usernetes.runner import UsernetesRunner


def main(args, _):
    runner = UsernetesRunner(workdir=args.workdir)
    runner.wait_for_workers(args.worker_count)
    # We assume that new workers need a sync
    runner.sync_external_ip()



import hashlib
import os
import socket

import usernetes.utils as utils
from usernetes.logger import logger


class ComposeConfig:
    """
    A compose config will help to prepare for the execution of docker compose!
    """

    def __init__(
        self,
        container_engine=None,
        node_service_name="node",
        container_engine_type="docker",
        compose="compose",
        bypass=False,
    ):
        self._node_name = None
        self._node_subnet = None
        self.node_service_name = node_service_name
        self.container_engine = container_engine or "docker"
        self.container_engine_type = container_engine_type or "docker"
        self.compose = compose or "compose"
        self.bypass = bypass

    @property
    def hostname(self):
        return socket.gethostname()

    def check(self):
        """
        Wrapper to check_preflight.
        """
        # We assume the context is in the pwd
        pwd = os.getcwd()
        print(f"Checking for Usernetes in {pwd}")

        # Since this is complex, we just wrap the script
        script = os.path.join(pwd, "Makefile.d", "check-preflight.sh")

        # Ensure to export envars the script needs
        envars = {
            "CONTAINER_ENGINE": self.container_engine,
            "CONTAINER_ENGINE_TYPE": self.container_engine_type,
        }

        # preflight checks must pass
        result = utils.run_command(["/bin/bash", script], envars=envars, stream=True)
        if result["return_code"] != 0:
            raise ValueError(f"Issue with preflight return code {result['return_code']}")

    def set_build_environment(self):
        """
        Export envars to the environment.
        """
        for k, v in self.envars.items():
            os.environ[k] = v
            os.putenv(k, v)

    @property
    def envars(self):
        """
        Return environment variables for interactive shell and other
        interactions.

        NODE_SHELL := $(COMPOSE) exec \
	    -e U7S_HOST_IP=$(U7S_HOST_IP) \
	    -e U7S_NODE_NAME=$(U7S_NODE_NAME) \
    	-e U7S_NODE_SUBNET=$(U7S_NODE_SUBNET) \
	    -e U7S_NODE_IP=$(U7S_NODE_IP) \
    	$(NODE_SERVICE_NAME)
        """
        return {
            "U7S_HOST_IP": self.usernetes_node_ip,
            "U7S_NODE_NAME": self.usernetes_node_name,
            "U7S_NODE_SUBNET": self.usernetes_node_subnet,
            "U7S_NODE_IP": self.usernetes_node_ip,
        }

    @property
    def node_name(self):
        if self._node_name is not None:
            return self._node_name
        self._node_name = f"u7s-{self.hostname}"
        return self.node_name

    @property
    def node_subnet(self):
        """
        Calculate the node subnet.
        """
        if self._node_subnet is not None:
            return self._node_subnet

        # Take the first two digits of the hexdigest
        hasher = hashlib.sha256()

        # We add a newline to mimic the command line variant
        # and get the same digest - otherwise they are different
        # NODE_SUBNET_ID=$((16#$(echo "${HOSTNAME}" | sha256sum | head -c2)))
        hasher.update((self.hostname + "\n").encode("utf-8"))
        digest = hasher.hexdigest()

        # Express the prefix in base 16 (e.g., $((16#<prefix>)))
        prefix = int(digest[0:2], 16)
        self._node_subnet = f"10.100.{prefix}.0/24"
        return self.node_subnet

    @property
    def usernetes_host_ip(self):
        """
        U7S_HOST_IP is the IP address of the physical host.
        Accessible from other hosts (convenience function)
        """
        return self.host_ip

    @property
    def usernetes_node_name(self):
        """
        U7S_NODE_NAME is the host name of the Kubernetes node running in Rootless Docker.
        Not accessible from other hosts.
        """
        return self.node_name

    @property
    def usernetes_node_subnet(self):
        """
        U7S_NODE_SUBNET is the subnet of the Kubernetes node running in Rootless Docker.
        Not accessible from other hosts.
        """
        return self.node_subnet

    @property
    def usernetes_node_ip(self):
        """
        # U7S_NODE_IP is the IP address of the Kubernetes node running in Rootless Docker.
        # Not accessible from other hosts.

        $(subset .0/24,.100,$(U7S_NODE_SUBNET))
        """
        return self.node_subnet.replace(".0/24", ".100")

    @property
    def host_ip(self):
        """
        Gets the host IP address
        """
        return socket.gethostbyname(self.hostname)



# These functions for environment and attributes assume
# being in a flux instance, and also assume it is created
# in the user temporary directory.
import os

import usernetes.instance as iutils


def attributes(contenders):
    """
    Return unique (or default) attributes.
    """
    contenders = list(set(contenders))
    if not contenders:
        contenders = ["kubeconfig", "workdir"]
    return contenders


def attr_main(args, _):
    """
    Get (and print) one or more attributes
    """
    instance = iutils.InstanceAttributes()
    for name in attributes(args.attributes):
        attribute = getattr(instance, name, None)
        if not attribute:
            continue
        print(attribute)


def env_main(args, _):
    """
    Get (and print) one or more envars
    """
    instance = iutils.InstanceAttributes()
    attribute_set = attributes(args.attributes)
    if "kubeconfig" in attribute_set:
        print(f"export KUBECONFIG={instance.kubeconfig}")
    if "workdir" in args.attributes:
        print(f"export USERNETES_WORKDIR={instance.workdir}")



# These functions for environment and attributes assume
# being in a flux instance, and also assume it is created
# in the user temporary directory.
import os
import tempfile

import usernetes.utils as utils


class InstanceAttributes:
    def __init__(self):
        self.jobid = get_jobid()
        self.uid = f"usernetes-{self.jobid.lower()}"
        self.root = os.path.join(tempfile.gettempdir(), self.uid)

    @property
    def kubeconfig(self):
        return os.path.join(self.root, "kubeconfig")

    @property
    def workdir(self):
        return self.root


def get_jobid():
    """
    Get the job id, first from an attribute then environ.

    This assumes flux, but arguably we can add other managers
    or they can export a similar identifier.
    """
    # This only works in batch (not run/submit)
    result = utils.run_command(["flux", "getattr", "jobid"])
    if result["return_code"] == 0 and result["message"] is not None:
        jobid = result["message"].strip()
    else:
        # Run/submit should have the envar
        jobid = os.environ.get("FLUX_JOB_ID")
    if jobid is None:
        # Fall back to Slurm
        jobid = os.environ.get("SLURM_JOB_ID")
    if jobid is None:
        raise ValueError("Cannot derive jobid to interact with.")
    return jobid



__version__ = "0.0.1"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsoch@users.noreply.github.com"
NAME = "usernetes"
PACKAGE_URL = "https://github.com/converged-computing/usernetes-python"
KEYWORDS = "cluster, orchestration, user-space kubernetes, kubernetes, compose"
DESCRIPTION = "Python wrapper for user-space Kubernetes 'usernetes'"
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = []
TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)
INSTALL_REQUIRES_ALL = TESTS_REQUIRES



#!/usr/bin/env python

import argparse
import os
import sys

import usernetes
import usernetes.defaults as defaults
from usernetes.logger import setup_logger


def get_parser():
    parser = argparse.ArgumentParser(
        description="Usernetes Python",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--develop",
        help="Don't wrap main in a try except (allow error to come through)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description="actions",
        dest="command",
    )
    subparsers.add_parser("version", description="show software version")
    start_worker = subparsers.add_parser(
        "start-worker",
        formatter_class=argparse.RawTextHelpFormatter,
        description="start user-space Kubernetes worker (akin to 'up')",
    )
    start_control = subparsers.add_parser(
        "start-control-plane",
        formatter_class=argparse.RawTextHelpFormatter,
        description="start user-space Kubernetes control plane (akin to 'up')",
    )
    start_control.add_argument(
        "--serial",
        help="Serial execution mode (do not wait for workers to come up)",
        action="store_true",
        default=False,
    )
    down = subparsers.add_parser(
        "down",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Bring down a node",
    )
    wait = subparsers.add_parser(
        "wait-workers",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Wait for workers and sync ip address when ready.",
    )
    # Env and attributes are obtainable from the instance.
    # E.g., we assume we can run flux getattr and similar.
    attribute = subparsers.add_parser(
        "attr",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Derive attributes",
    )
    envars = subparsers.add_parser(
        "env",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Derive environment variables",
    )
    for command in [attribute, envars]:
        command.add_argument(
            "attributes",
            action="append",
            default=[],
            help="Return (by default) kubeconfig and workdir",
        )
    clean = subparsers.add_parser(
        "clean",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Clean the worker cache (to prepare for another job)",
    )
    clean.add_argument(
        "--keep-node",
        dest="keep_node",
        help="Do not remove the node (requires keeping the workdir)",
    )
    for command in [start_control, clean]:
        command.add_argument(
            "--all",
            dest="clean_all",
            help="Also remove the entire usernetes root (not just worker cache)",
            action="store_true",
            default=False,
        )
    for command in [start_worker, start_control, wait, down, clean]:
        command.add_argument(
            "--workdir",
            help="working directory with docker-compose.yaml",
        )
    for command in [start_control, wait]:
        command.add_argument(
            "--worker-count",
            help="worker count (not including control plane)",
        )
    return parser


def run_usernetes():
    """
    this is the main entrypoint.
    """
    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """
        version = usernetes.__version__

        print("\nUsernetes Python v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(usernetes.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # Here we can assume instantiated to get args
    if args.command in ["start-worker", "start-control-plane"]:
        from .start import main
    elif args.command == "down":
        from .down import main
    elif args.command == "wait-workers":
        from .wait import main
    elif args.command == "clean":
        from .clean import main
    elif args.command == "attr":
        from .environment import attr_main as main
    elif args.command == "env":
        from .environment import env_main as main

    # Develop mode, akin to commenting out the try/except below
    if args.develop:
        return main(args, extra)
    try:
        main(args, extra)
    except:
        help(1)


if __name__ == "__main__":
    run_usernetes()



from usernetes.runner import UsernetesRunner


def main(args, _):
    runner = UsernetesRunner(workdir=args.workdir)
    if args.command == "start-worker":
        runner.start_worker()
    else:
        if not args.worker_count:
            raise ValueError("A --worker-count is required.")
        runner.start_control_plane(args.worker_count, serial=args.serial)
